#!/usr/bin/env python3
# © Copyright CERN, Riga Technical University and University of Liverpool 2020.
# All rights not expressly granted are reserved.
#
# This file is part of hev-sw.
#
# hev-sw is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free
# Software Foundation, either version 3 of the Licence, or (at your option)
# any later version.
#
# hev-sw is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with hev-sw. If not, see <http://www.gnu.org/licenses/> .
#
# The authors would like to acknowledge the much appreciated support
# of all those involved with the High Energy Ventilator project
# (https://hev.web.cern.ch/).


# client side for the data server to manage comms between UIs and LLI
#
# Author: Dónal Murray <donal.murray@cern.ch>

import asyncio
import time
import json
import threading
from typing import List, Dict, Union
from CommsCommon import PAYLOAD_TYPE
import logging

logging.basicConfig(
    level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s"
)


class HEVPacketError(Exception):
    pass


class HEVClient(object):
    def __init__(self, polling=True):
        super().__init__()
        # alarms latch
        self._alarms = []

        # Last packet of each type
        self._last_packet = {
            "PERSONAL": None,
            "DATA": None,
            "READBACK": None,
            "CYCLE": None,
            "TARGET": None,
            "LOGMSG": None,
            "THRESHOLDS": None,
        }

        # lock for the last packet storage
        self._lock = threading.Lock()

        # start polling in another thread unless told otherwise
        self._polling = polling
        if self._polling:
            self.start_polling()

    def start_polling(self):
        """start worker thread to get data in the background"""
        worker = threading.Thread(target=self.start_client, daemon=True)
        worker.start()

    def start_client(self) -> None:
        """synchronous wrapper for polling()"""
        asyncio.run(self.polling())

    async def polling(self) -> None:
        """open persistent connection with server"""
        writer = None
        while True:
            try:
                reader, writer = await asyncio.open_connection("127.0.0.1", 54320)

                while self._polling:
                    try:
                        # grab data from the socket as soon as it is available
                        data = await reader.readuntil(separator=b"\0")
                        # snip off nullbyte and parse json
                        data = data[:-1]
                        payload = json.loads(data.decode("utf-8"))
                        payload_type = payload["type"]

                        # check that it's valid
                        if payload_type == "keepalive":
                            # drop silently - without it the hevserver can go down unnoticed
                            continue
                        elif payload_type in self._last_packet:
                            # store it in the right place and pass it on
                            with self._lock:
                                self._last_packet[payload_type] = payload[payload_type]
                        elif payload_type in [pt.name for pt in PAYLOAD_TYPE]:
                            # valid payload type but unimplemented in hevclient
                            # passing lets us still use alarm information vs continue
                            pass
                        else:
                            # invalid packet - something seriously wrong with the hevserver
                            raise HEVPacketError("Invalid broadcast type")

                        self._alarms = payload["alarms"]
                        self.get_updates(payload)  # callback function to be overridden
                    except json.decoder.JSONDecodeError:
                        logging.warning(f"Could not decode packet: {data}")
                    except KeyError:
                        raise

            except ConnectionRefusedError as e:
                logging.error(str(e) + " - is the microcontroller running?")
                await asyncio.sleep(2)
            except Exception as e:
                # warn and reopen connection
                logging.error(e)
                await asyncio.sleep(2)
            finally:
                # close connection
                if writer is not None:
                    writer.close()
                    await writer.wait_closed()

    def get_updates(self, payload) -> None:
        """Overrideable function called after receiving data from the socket, with that data as an argument"""
        logging.debug(payload)
        pass

    async def _send_request(
        self,
        reqtype,
        cmdtype: str = None,
        cmd: str = None,
        param: str = None,
        alarm: str = None,
        personal: str = None,
    ) -> bool:
        # open connection and send packet
        reader, writer = await asyncio.open_connection("127.0.0.1", 54321)
        payload = {"type": reqtype}
        if reqtype == "CMD":
            payload["cmdtype"] = cmdtype
            payload["cmd"] = cmd
            payload["param"] = param
        elif reqtype == "ALARM":
            payload["alarm_type"] = alarm["alarm_type"]
            payload["alarm_code"] = alarm["alarm_code"]
            payload["param"] = param
        elif reqtype == "PERSONAL":
            payload["name"] = personal["name"]
            payload["age"] = int(personal["age"])
            payload["sex"] = personal["sex"]
            payload["height"] = int(personal["height"])
            payload["weight"] = int(personal["weight"])
        else:
            raise HEVPacketError("Invalid packet type")

        logging.info(payload)
        packet = json.dumps(payload).encode() + b"\0"

        writer.write(packet)
        await writer.drain()

        # wait for acknowledge
        data = await reader.readuntil(separator=b"\0")
        data = data[:-1]
        try:
            data = json.loads(data.decode("utf-8"))
        except json.decoder.JSONDecodeError:
            logging.warning(f"Could not decode packet: {data}")

        # close connection to free up socket
        writer.close()
        await writer.wait_closed()

        # check that acknowledge is meaningful
        if data["type"] == "ack":
            logging.info(f"Request type {reqtype} sent successfully")
            return True
        else:
            logging.warning(f"Request type {reqtype} failed")
            return False

    def send_cmd(self, cmdtype: str, cmd: str, param: Union[float, int] = None) -> bool:
        # send a cmd and wait to see if it's valid
        # print(cmdtype, cmd, param)
        try:
            return asyncio.run(
                self._send_request("CMD", cmdtype=cmdtype, cmd=cmd, param=param)
            )
        except ConnectionRefusedError as error:
            logging.error(str(error) + " - is the microcontroller running?")

    def ack_alarm(self, alarm: str) -> bool:
        # acknowledge alarm to remove it from the hevserver list
        return asyncio.run(self._send_request("ALARM", alarm=alarm))

    # def send_personal(self, personal: Dict[str, str]=None ) -> bool:
    def send_personal(self, personal: str) -> bool:
        # send personal information
        return asyncio.run(self._send_request("PERSONAL", personal=personal))

    # Getters to check last value of each datatype - debugging only
    def get_values(self) -> Dict:
        return self._last_packet["DATA"]

    def get_readback(self) -> Dict:
        return self._last_packet["READBACK"]

    def get_cycle(self) -> Dict:
        return self._last_packet["CYCLE"]

    def get_personal(self) -> Dict:
        return self._last_packet["PERSONAL"]

    def get_logmsg(self) -> Dict:
        return self._last_packet["LOGMSG"]

    def get_target(self) -> Dict:
        return self._last_packet["TARGET"]

    def get_thresholds(self) -> Dict:
        return self._last_packet["THRESHOLDS"]

    def get_alarms(self) -> List[str]:
        return self._alarms


if __name__ == "__main__":
    # example implementation
    # just import hevclient and do something like the following
    hevclient = HEVClient()

    time.sleep(2)
    print(hevclient.send_cmd("GENERAL", "START"))
    # Play with sensor values and alarms
    for i in range(20):
        values = hevclient.get_values()  # returns a dict or None
        alarms = hevclient.get_alarms()  # returns a list of alarms currently ongoing
        print(values)
        if values is None:
            i = i + 1 if i > 0 else 0
        else:
            print(f"Values: {json.dumps(values, indent=4)}")
            print(f"Alarms: {alarms}")
        time.sleep(1)

    # acknowledge the oldest alarm
    try:
        hevclient.ack_alarm(alarms[0])  # blindly assume we have one after 40s
    except:
        logging.info("No alarms received")

    time.sleep(2)
    print(f"Alarms: {hevclient.get_alarms()}")

    # set a timeout
    hevclient.send_cmd("SET_DURATION", "INHALE", 1111)

    # check for the readback
    for i in range(10):
        print(f"Readback: {json.dumps(hevclient.get_readback(), indent=4)}")
        print(f"Alarms: {hevclient.get_alarms()}")
        time.sleep(1)

    print(hevclient.send_cmd("GENERAL", "STOP"))
